# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implements Deductive Database (DD)."""

from __future__ import annotations
from typing import Any, Callable, Generator

from collections import defaultdict
import time


import newclid.geometry as gm
import newclid.graph as gh
import newclid.graph_utils as utils
import newclid.numericals as nm
import newclid.problem as pr
from newclid.problem import Dependency, EmptyDependency


def intersect1(set1: set[Any], set2: set[Any]) -> Any:
    for x in set1:
        if x in set2:
            return x
    return None


def diff_point(line: gm.Line, a: gm.Point) -> gm.Point:
    for x in line.neighbors(gm.Point):
        if x != a:
            return x
    return None


def match_eqratio_eqratio_eqratio(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqratio a b c d m n p q, eqratio c d e f p q r u => eqratio a b e f m n r u."""
    for m1 in g.type2nodes[gm.Value]:
        for m2 in g.type2nodes[gm.Value]:
            rats1 = []
            for rat in m1.neighbors(gm.Ratio):
                l1, l2 = rat.lengths
                if l1 is None or l2 is None:
                    continue
                rats1.append((l1, l2))

            rats2 = []
            for rat in m2.neighbors(gm.Ratio):
                l1, l2 = rat.lengths
                if l1 is None or l2 is None:
                    continue
                rats2.append((l1, l2))

            pairs = []
            for (l1, l2), (l3, l4) in utils.cross(rats1, rats2):
                if l2 == l3:
                    pairs.append((l1, l2, l4))

            for (l1, l12, l2), (l3, l34, l4) in utils.comb2(pairs):
                if (l1, l12, l2) == (l3, l34, l4):
                    continue
                if l1 == l2 or l3 == l4:
                    continue
                if l1 == l12 or l12 == l2 or l3 == l34 or l4 == l34:
                    continue
                # d12 - d1 = d34 - d3 = m1
                # d2 - d12 = d4 - d34 = m2
                # => d2 - d1 = d4 - d3 (= m1+m2)
                a, b = g.two_points_of_length(l1)
                c, d = g.two_points_of_length(l12)
                m, n = g.two_points_of_length(l3)
                p, q = g.two_points_of_length(l34)
                # eqangle a b c d m n p q
                e, f = g.two_points_of_length(l2)
                r, u = g.two_points_of_length(l4)
                yield dict(zip("abcdefmnpqru", [a, b, c, d, e, f, m, n, p, q, r, u]))


def match_eqangle_eqangle_eqangle(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle a b c d m n p q, eqangle c d e f p q r u => eqangle a b e f m n r u."""
    for m1 in g.type2nodes[gm.Measure]:
        for m2 in g.type2nodes[gm.Measure]:
            angs1 = []
            for ang in m1.neighbors(gm.Angle):
                d1, d2 = ang.directions
                if d1 is None or d2 is None:
                    continue
                angs1.append((d1, d2))

            angs2 = []
            for ang in m2.neighbors(gm.Angle):
                d1, d2 = ang.directions
                if d1 is None or d2 is None:
                    continue
                angs2.append((d1, d2))

            pairs = []
            for (d1, d2), (d3, d4) in utils.cross(angs1, angs2):
                if d2 == d3:
                    pairs.append((d1, d2, d4))

            for (d1, d12, d2), (d3, d34, d4) in utils.comb2(pairs):
                if (d1, d12, d2) == (d3, d34, d4):
                    continue
                if d1 == d2 or d3 == d4:
                    continue
                if d1 == d12 or d12 == d2 or d3 == d34 or d4 == d34:
                    continue
                # d12 - d1 = d34 - d3 = m1
                # d2 - d12 = d4 - d34 = m2
                # => d2 - d1 = d4 - d3
                a, b = g.two_points_on_direction(d1)
                c, d = g.two_points_on_direction(d12)
                m, n = g.two_points_on_direction(d3)
                p, q = g.two_points_on_direction(d34)
                # eqangle a b c d m n p q
                e, f = g.two_points_on_direction(d2)
                r, u = g.two_points_on_direction(d4)
                yield dict(zip("abcdefmnpqru", [a, b, c, d, e, f, m, n, p, q, r, u]))


def match_perp_perp_npara_eqangle(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match perp A B C D, perp E F G H, npara A B E F => eqangle A B E F C D G H."""
    dpairs = []
    for ang in g.vhalfpi.neighbors(gm.Angle):
        d1, d2 = ang.directions
        if d1 is None or d2 is None:
            continue
        dpairs.append((d1, d2))

    for (d1, d2), (d3, d4) in utils.comb2(dpairs):
        a, b = g.two_points_on_direction(d1)
        c, d = g.two_points_on_direction(d2)
        m, n = g.two_points_on_direction(d3)
        p, q = g.two_points_on_direction(d4)
        if g.check_npara([a, b, m, n]):
            if ({a, b}, {c, d}) == ({m, n}, {p, q}):
                continue
            if ({a, b}, {c, d}) == ({p, q}, {m, n}):
                continue

            yield dict(zip("ABCDEFGH", [a, b, c, d, m, n, p, q]))


def match_circle_coll_eqangle_midp(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match circle O A B C, coll M B C, eqangle A B A C O B O M => midp M B C."""
    for p, a, b, c in g.all_circles():
        ab = g._get_line(a, b)
        if ab is None:
            continue
        if ab.val is None:
            continue
        ac = g._get_line(a, c)
        if ac is None:
            continue
        if ac.val is None:
            continue
        pb = g._get_line(p, b)
        if pb is None:
            continue
        if pb.val is None:
            continue

        bc = g._get_line(b, c)
        if bc is None:
            continue
        bc_points = bc.neighbors(gm.Point, return_set=True)

        anga, _ = g._get_angle(ab.val, ac.val)

        for angp in pb.val.neighbors(gm.Angle):
            if not g.is_equal(anga, angp):
                continue

            _, d = angp.directions
            for line_neighbor in d.neighbors(gm.Line):
                l_points = line_neighbor.neighbors(gm.Point, return_set=True)
                m = intersect1(bc_points, l_points)
                if m is not None:
                    yield dict(zip("ABCMO", [a, b, c, m, p]))


def match_midp_perp_cong(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match midp M A B, perp O M A B => cong O A O B."""
    for m, a, b in g.all_midps():
        ab = g._get_line(a, b)
        for line_neighbor in m.neighbors(gm.Line):
            if g.check_perpl(line_neighbor, ab):
                for o in line_neighbor.neighbors(gm.Point):
                    if o != m:
                        yield dict(zip("ABMO", [a, b, m, o]))


def match_cyclic_eqangle_cong(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match cyclic A B C P Q R, eqangle C A C B R P R Q => cong A B P Q."""
    for c in g.type2nodes[gm.Circle]:
        ps = c.neighbors(gm.Point)
        for (a, b, c), (x, y, z) in utils.comb2(list(utils.perm3(ps))):
            if {a, b, c} == {x, y, z}:
                continue
            if g.check_eqangle([c, a, c, b, z, x, z, y]):
                yield dict(zip("ABCPQR", [a, b, c, x, y, z]))


def match_circle_eqangle_perp(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match circle O A B C, eqangle A X A B C A C B => perp O A A X."""
    for p, a, b, c in g.all_circles():
        ca = g._get_line(c, a)
        if ca is None:
            continue
        cb = g._get_line(c, b)
        if cb is None:
            continue
        ab = g._get_line(a, b)
        if ab is None:
            continue

        if ca.val is None:
            continue
        if cb.val is None:
            continue
        if ab.val is None:
            continue

        c_ang, _ = g._get_angle(cb.val, ca.val)
        if c_ang is None:
            continue

        for ang in ab.val.neighbors(gm.Angle):
            if g.is_equal(ang, c_ang):
                _, d = ang.directions
                for line_neighbor in d.neighbors(gm.Line):
                    if a not in line_neighbor.neighbors(gm.Point):
                        continue
                    x = diff_point(line_neighbor, a)
                    if x is None:
                        continue
                    yield dict(zip("OABCX", [p, a, b, c, x]))
                break


def match_circle_perp_eqangle(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match circle O A B C, perp O A A X => eqangle A X A B C A C B."""
    for p, a, b, c in g.all_circles():
        pa = g._get_line(p, a)
        if pa is None:
            continue
        if pa.val is None:
            continue
        for line_neighbor in a.neighbors(gm.Line):
            if g.check_perpl(pa, line_neighbor):
                x = diff_point(line_neighbor, a)
                if x is not None:
                    yield dict(zip("OABCX", [p, a, b, c, x]))


def match_perp_perp_ncoll_para(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match perp A B C D, perp C D E F, ncoll A B E => para A B E F."""
    d2d = defaultdict(list)
    for ang in g.vhalfpi.neighbors(gm.Angle):
        d1, d2 = ang.directions
        if d1 is None or d2 is None:
            continue
        d2d[d1] += [d2]
        d2d[d2] += [d1]

    for x, ys in d2d.items():
        if len(ys) < 2:
            continue
        c, d = g.two_points_on_direction(x)
        for y1, y2 in utils.comb2(ys):
            a, b = g.two_points_on_direction(y1)
            e, f = g.two_points_on_direction(y2)
            if nm.check_ncoll([a.num, b.num, e.num]):
                yield dict(zip("ABCDEF", [a, b, c, d, e, f]))


def match_eqangle6_ncoll_cong(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle6 A O A B B A B O, ncoll O A B => cong O A O B."""
    for a in g.type2nodes[gm.Point]:
        for b, c in utils.comb2(g.type2nodes[gm.Point]):
            if a == b or a == c:
                continue
            if g.check_eqangle([b, a, b, c, c, b, c, a]):
                if g.check_ncoll([a, b, c]):
                    yield dict(zip("OAB", [a, b, c]))


def match_eqangle_perp_perp(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle A B P Q C D U V, perp P Q U V => perp A B C D."""
    for ang in g.vhalfpi.neighbors(gm.Angle):
        # d1 perp d2
        d1, d2 = ang.directions
        if d1 is None or d2 is None:
            continue
        for d3, d4 in utils.comb2(g.type2nodes[gm.Direction]):
            if d1 == d3 or d2 == d4:
                continue
            # if d1 - d3 = d2 - d4 => d3 perp d4
            a13, a31 = g._get_angle(d1, d3)
            a24, a42 = g._get_angle(d2, d4)
            if a13 is None or a31 is None or a24 is None or a42 is None:
                continue
            if g.is_equal(a13, a24) and g.is_equal(a31, a42):
                a, b = g.two_points_on_direction(d1)
                c, d = g.two_points_on_direction(d2)
                m, n = g.two_points_on_direction(d3)
                p, q = g.two_points_on_direction(d4)
                yield dict(zip("ABCDPQUV", [m, n, p, q, a, b, c, d]))


def match_eqangle_ncoll_cyclic(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle6 P A P B Q A Q B, ncoll P Q A B => cyclic A B P Q."""
    for l1, l2, l3, l4 in g.all_eqangles_distinct_linepairss():
        if len(set([l1, l2, l3, l4])) < 4:
            continue  # they all must be distinct.

        p1s = l1.neighbors(gm.Point, return_set=True)
        p2s = l2.neighbors(gm.Point, return_set=True)
        p3s = l3.neighbors(gm.Point, return_set=True)
        p4s = l4.neighbors(gm.Point, return_set=True)

        p = intersect1(p1s, p2s)
        if not p:
            continue
        q = intersect1(p3s, p4s)
        if not q:
            continue
        a = intersect1(p1s, p3s)
        if not a:
            continue
        b = intersect1(p2s, p4s)
        if not b:
            continue
        if len(set([a, b, p, q])) < 4:
            continue

        if not g.check_ncoll([a, b, p, q]):
            continue

        yield dict(zip("ABPQ", [a, b, p, q]))


def match_eqangle_para(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle A B P Q C D P Q => para A B C D."""
    for measure in g.type2nodes[gm.Measure]:
        angs = measure.neighbors(gm.Angle)
        d12, d21 = defaultdict(list), defaultdict(list)
        for ang in angs:
            d1, d2 = ang.directions
            if d1 is None or d2 is None:
                continue
            d12[d1].append(d2)
            d21[d2].append(d1)

        for d1, d2s in d12.items():
            a, b = g.two_points_on_direction(d1)
            for d2, d3 in utils.comb2(d2s):
                c, d = g.two_points_on_direction(d2)
                e, f = g.two_points_on_direction(d3)
                yield dict(zip("ABCDPQ", [c, d, e, f, a, b]))


def match_cyclic_eqangle(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match cyclic A B P Q => eqangle P A P B Q A Q B."""
    record = set()
    for a, b, c, d in g_matcher("cyclic"):
        if (a, b, c, d) in record:
            continue
        record.add((a, b, c, d))
        record.add((a, b, d, c))
        record.add((b, a, c, d))
        record.add((b, a, d, c))
        yield dict(zip("ABPQ", [a, b, c, d]))


def rotate_simtri(
    a: gm.Point, b: gm.Point, c: gm.Point, x: gm.Point, y: gm.Point, z: gm.Point
) -> Generator[tuple[gm.Point, ...], None, None]:
    """Rotate points around for similar triangle predicates."""
    yield (z, y, x, c, b, a)
    for p in [
        (b, c, a, y, z, x),
        (c, a, b, z, x, y),
        (x, y, z, a, b, c),
        (y, z, x, b, c, a),
        (z, x, y, c, a, b),
    ]:
        yield p
        yield p[::-1]


def match_cong_cong_cong_cyclic(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match cong O A O B, cong O B O C, cong O C O D => cyclic A B C D."""
    for lenght in g.type2nodes[gm.Length]:
        p2p = defaultdict(list)
        for s in lenght.neighbors(gm.Segment):
            a, b = s.points
            p2p[a].append(b)
            p2p[b].append(a)

        for p, ps in p2p.items():
            if len(ps) >= 4:
                for a, b, c, d in utils.comb4(ps):
                    yield dict(zip("OABCD", [p, a, b, c, d]))


def match_cong_cong_cong_ncoll_contri(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match cong A B P Q, cong B C Q R, cong C A R P, ncoll A B C => contri* A B C P Q R."""
    record = set()
    for a, b, p, q in g_matcher("cong"):
        for c in g.type2nodes[gm.Point]:
            for r in g.type2nodes[gm.Point]:
                if any([x in record for x in rotate_simtri(a, b, c, p, q, r)]):
                    continue
                if not g.check_ncoll([a, b, c]):
                    continue
                if g.check_cong([b, c, q, r]) and g.check_cong([c, a, r, p]):
                    record.add((a, b, c, p, q, r))
                    yield dict(zip("ABCPQR", [a, b, c, p, q, r]))


def match_cong_cong_eqangle6_ncoll_contri(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match cong A B P Q, cong B C Q R, eqangle6 B A B C Q P Q R, ncoll A B C => contri* A B C P Q R."""
    record = set()
    for a, b, p, q in g_matcher("cong"):
        for c in g.type2nodes[gm.Point]:
            if c in (a, b):
                continue
            for r in g.type2nodes[gm.Point]:
                if r in (p, q):
                    continue

                in_record = False
                for x in [
                    (c, b, a, r, q, p),
                    (p, q, r, a, b, c),
                    (r, q, p, c, b, a),
                ]:
                    if x in record:
                        in_record = True
                        break

                if in_record:
                    continue

                if not g.check_cong([b, c, q, r]):
                    continue
                if not g.check_ncoll([a, b, c]):
                    continue

                if nm.same_clock(a.num, b.num, c.num, p.num, q.num, r.num):
                    if g.check_eqangle([b, a, b, c, q, p, q, r]):
                        record.add((a, b, c, p, q, r))
                        yield dict(zip("ABCPQR", [a, b, c, p, q, r]))
                else:
                    if g.check_eqangle([b, a, b, c, q, r, q, p]):
                        record.add((a, b, c, p, q, r))
                        yield dict(zip("ABCPQR", [a, b, c, p, q, r]))


def match_eqratio6_eqangle6_ncoll_simtri(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqratio6 B A B C Q P Q R, eqratio6 C A C B R P R Q, ncoll A B C => simtri* A B C P Q R."""
    enums = g_matcher("eqratio6")

    record = set()
    for b, a, b, c, q, p, q, r in enums:
        if (a, b, c) == (p, q, r):
            continue
        if any([x in record for x in rotate_simtri(a, b, c, p, q, r)]):
            continue
        if not g.check_ncoll([a, b, c]):
            continue

        if nm.same_clock(a.num, b.num, c.num, p.num, q.num, r.num):
            if g.check_eqangle([b, a, b, c, q, p, q, r]):
                record.add((a, b, c, p, q, r))
                yield dict(zip("ABCPQR", [a, b, c, p, q, r]))
        elif g.check_eqangle([b, a, b, c, q, r, q, p]):
            record.add((a, b, c, p, q, r))
            yield dict(zip("ABCPQR", [a, b, c, p, q, r]))


def match_eqangle6_eqangle6_ncoll_simtri(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle6 B A B C Q P Q R, eqangle6 C A C B R P R Q, ncoll A B C => simtri A B C P Q R."""
    enums = g_matcher("eqangle6")

    record = set()
    for b, a, b, c, q, p, q, r in enums:
        if (a, b, c) == (p, q, r):
            continue
        if any([x in record for x in rotate_simtri(a, b, c, p, q, r)]):
            continue
        if not g.check_eqangle([c, a, c, b, r, p, r, q]):
            continue
        if not g.check_ncoll([a, b, c]):
            continue

        mapping = dict(zip("ABCPQR", [a, b, c, p, q, r]))
        record.add((a, b, c, p, q, r))
        yield mapping


def match_eqratio6_eqratio6_ncoll_simtri(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqratio6 B A B C Q P Q R, eqratio6 C A C B R P R Q, ncoll A B C => simtri* A B C P Q R."""
    enums = g_matcher("eqratio6")

    record = set()
    for b, a, b, c, q, p, q, r in enums:
        if (a, b, c) == (p, q, r):
            continue
        if any([x in record for x in rotate_simtri(a, b, c, p, q, r)]):
            continue
        if not g.check_eqratio([c, a, c, b, r, p, r, q]):
            continue
        if not g.check_ncoll([a, b, c]):
            continue

        mapping = dict(zip("ABCPQR", [a, b, c, p, q, r]))
        record.add((a, b, c, p, q, r))
        yield mapping


def match_eqangle6_eqangle6_ncoll_simtri2(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle6 B A B C Q R Q P, eqangle6 C A C B R Q R P, ncoll A B C => simtri2 A B C P Q R."""
    enums = g_matcher("eqangle6")

    record = set()
    for b, a, b, c, q, r, q, p in enums:
        if (a, b, c) == (p, q, r):
            continue
        if any([x in record for x in rotate_simtri(a, b, c, p, q, r)]):
            continue
        if not g.check_eqangle([c, a, c, b, r, q, r, p]):
            continue
        if not g.check_ncoll([a, b, c]):
            continue

        mapping = dict(zip("ABCPQR", [a, b, c, p, q, r]))
        record.add((a, b, c, p, q, r))
        yield mapping


def rotate_contri(
    a: gm.Point, b: gm.Point, c: gm.Point, x: gm.Point, y: gm.Point, z: gm.Point
) -> Generator[tuple[gm.Point, ...], None, None]:
    for p in [(b, a, c, y, x, z), (x, y, z, a, b, c), (y, x, z, b, a, c)]:
        yield p


def match_eqangle6_eqangle6_ncoll_cong_contri(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle6 B A B C Q P Q R, eqangle6 C A C B R P R Q, ncoll A B C, cong A B P Q => contri A B C P Q R."""
    enums = g_matcher("eqangle6")

    record = set()
    for b, a, b, c, q, p, q, r in enums:
        if not g.check_cong([a, b, p, q]):
            continue
        if (a, b, c) == (p, q, r):
            continue
        if any([x in record for x in rotate_contri(a, b, c, p, q, r)]):
            continue
        if not g.check_eqangle([c, a, c, b, r, p, r, q]):
            continue

        if not g.check_ncoll([a, b, c]):
            continue

        mapping = dict(zip("ABCPQR", [a, b, c, p, q, r]))
        record.add((a, b, c, p, q, r))
        yield mapping


def match_eqratio6_eqratio6_ncoll_cong_contri(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqratio6 B A B C Q P Q R, eqratio6 C A C B R P R Q, ncoll A B C, cong A B P Q => contri* A B C P Q R."""
    enums = g_matcher("eqratio6")

    record = set()
    for b, a, b, c, q, p, q, r in enums:
        if not g.check_cong([a, b, p, q]):
            continue
        if (a, b, c) == (p, q, r):
            continue
        if any([x in record for x in rotate_contri(a, b, c, p, q, r)]):
            continue
        if not g.check_eqratio([c, a, c, b, r, p, r, q]):
            continue

        if not g.check_ncoll([a, b, c]):
            continue

        mapping = dict(zip("ABCPQR", [a, b, c, p, q, r]))
        record.add((a, b, c, p, q, r))
        yield mapping


def match_eqangle6_eqangle6_ncoll_cong_contri2(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle6 B A B C Q R Q P, eqangle6 C A C B R Q R P, ncoll A B C, cong A B P Q => contri2 A B C P Q R."""
    enums = g_matcher("eqangle6")

    record = set()
    for b, a, b, c, q, r, q, p in enums:
        if not g.check_cong([a, b, p, q]):
            continue
        if (a, b, c) == (p, q, r):
            continue
        if any([x in record for x in rotate_contri(a, b, c, p, q, r)]):
            continue
        if not g.check_eqangle([c, a, c, b, r, q, r, p]):
            continue
        if not g.check_ncoll([a, b, c]):
            continue

        mapping = dict(zip("ABCPQR", [a, b, c, p, q, r]))
        record.add((a, b, c, p, q, r))
        yield mapping


def match_eqratio6_coll_ncoll_eqangle6(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqratio6 d b d c a b a c, coll d b c, ncoll a b c => eqangle6 a b a d a d a c."""
    records = set()
    for b, d, c in g_matcher("coll"):
        for a in g.all_points():
            if g.check_coll([a, b, c]):
                continue
            if (a, b, d, c) in records or (a, c, d, b) in records:
                continue
            records.add((a, b, d, c))

            if g.check_eqratio([d, b, d, c, a, b, a, c]):
                yield dict(zip("abcd", [a, b, c, d]))


def match_eqangle6_coll_ncoll_eqratio6(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle6 a b a d a d a c, coll d b c, ncoll a b c => eqratio6 d b d c a b a c."""
    records = set()
    for b, d, c in g_matcher("coll"):
        for a in g.all_points():
            if g.check_coll([a, b, c]):
                continue
            if (a, b, d, c) in records or (a, c, d, b) in records:
                continue
            records.add((a, b, d, c))

            if g.check_eqangle([a, b, a, d, a, d, a, c]):
                yield dict(zip("abcd", [a, b, c, d]))


def match_eqangle6_ncoll_cyclic(
    g: gh.Graph,
    g_matcher: Callable[[str], list[tuple[gm.Point, ...]]],
    theorem: pr.Theorem,
) -> Generator[dict[str, gm.Point], None, None]:
    """Match eqangle6 P A P B Q A Q B, ncoll P Q A B => cyclic A B P Q."""
    for a, b, a, c, x, y, x, z in g_matcher("eqangle6"):
        if (b, c) != (y, z) or a == x:
            continue
        if nm.check_ncoll([x.num for x in [a, b, c, x]]):
            yield dict(zip("ABPQ", [b, c, a, x]))


def match_all(name: str, g: gh.Graph) -> Generator[tuple[gm.Point, ...], None, None]:
    """Match all instances of a certain relation."""
    if name in ["ncoll", "npara", "nperp"]:
        return []
    if name == "coll":
        return g.all_colls()
    if name == "para":
        return g.all_paras()
    if name == "perp":
        return g.all_perps()
    if name == "cong":
        return g.all_congs()
    if name == "eqangle":
        return g.all_eqangles_8points()
    if name == "eqangle6":
        return g.all_eqangles_6points()
    if name == "eqratio":
        return g.all_eqratios_8points()
    if name == "eqratio6":
        return g.all_eqratios_6points()
    if name == "cyclic":
        return g.all_cyclics()
    if name == "midp":
        return g.all_midps()
    if name == "circle":
        return g.all_circles()
    raise ValueError(f"Unrecognize {name}")


def cache_match(
    graph: gh.Graph,
) -> Callable[[str], list[tuple[gm.Point, ...]]]:
    """Cache throughout one single BFS level."""
    cache = {}

    def match_fn(name: str) -> list[tuple[gm.Point, ...]]:
        if name in cache:
            return cache[name]

        result = list(match_all(name, graph))
        cache[name] = result
        return result

    return match_fn


def try_to_map(
    clause_enum: list[tuple[pr.Clause, list[tuple[gm.Point, ...]]]],
    mapping: dict[str, gm.Point],
) -> Generator[dict[str, gm.Point], None, None]:
    """Recursively try to match the remaining points given current mapping."""
    if not clause_enum:
        yield mapping
        return

    clause, enum = clause_enum[0]
    for points in enum:
        mpcpy = dict(mapping)

        fail = False
        for p, a in zip(points, clause.args):
            if a in mpcpy and mpcpy[a] != p or p in mpcpy and mpcpy[p] != a:
                fail = True
                break
            mpcpy[a] = p
            mpcpy[p] = a

        if fail:
            continue

        for m in try_to_map(clause_enum[1:], mpcpy):
            yield m


def match_generic(
    g: gh.Graph, cache: Callable[[str], list[tuple[gm.Point, ...]]], theorem: pr.Theorem
) -> Generator[dict[str, gm.Point], None, None]:
    """Match any generic rule that is not one of the above match_*() rules."""
    clause2enum = {}

    clauses = []
    numerical_checks = []
    for clause in theorem.premise:
        if clause.name in ["ncoll", "npara", "nperp", "sameside"]:
            numerical_checks.append(clause)
            continue

        enum = cache(clause.name)
        if len(enum) == 0:
            return 0

        clause2enum[clause] = enum
        clauses.append((len(set(clause.args)), clause))

    clauses = sorted(clauses, key=lambda x: x[0], reverse=True)
    _, clauses = zip(*clauses)

    for mapping in try_to_map([(c, clause2enum[c]) for c in clauses], {}):
        if not mapping:
            continue

        checks_ok = True
        for check in numerical_checks:
            args = [mapping[a] for a in check.args]
            if check.name == "ncoll":
                checks_ok = g.check_ncoll(args)
            elif check.name == "npara":
                checks_ok = g.check_npara(args)
            elif check.name == "nperp":
                checks_ok = g.check_nperp(args)
            elif check.name == "sameside":
                checks_ok = g.check_sameside(args)
            if not checks_ok:
                break
        if not checks_ok:
            continue

        yield mapping


BUILT_IN_FNS = {
    "cong_cong_cong_cyclic": match_cong_cong_cong_cyclic,
    "cong_cong_cong_ncoll_contri*": match_cong_cong_cong_ncoll_contri,
    "cong_cong_eqangle6_ncoll_contri*": match_cong_cong_eqangle6_ncoll_contri,
    "eqangle6_eqangle6_ncoll_simtri": match_eqangle6_eqangle6_ncoll_simtri,
    "eqangle6_eqangle6_ncoll_cong_contri": (match_eqangle6_eqangle6_ncoll_cong_contri),
    "eqangle6_eqangle6_ncoll_simtri2": match_eqangle6_eqangle6_ncoll_simtri2,
    "eqangle6_eqangle6_ncoll_cong_contri2": (
        match_eqangle6_eqangle6_ncoll_cong_contri2
    ),
    "eqratio6_eqratio6_ncoll_simtri*": match_eqratio6_eqratio6_ncoll_simtri,
    "eqratio6_eqratio6_ncoll_cong_contri*": (match_eqratio6_eqratio6_ncoll_cong_contri),
    "eqangle_para": match_eqangle_para,
    "eqangle_ncoll_cyclic": match_eqangle_ncoll_cyclic,
    "eqratio6_eqangle6_ncoll_simtri*": match_eqratio6_eqangle6_ncoll_simtri,
    "eqangle_perp_perp": match_eqangle_perp_perp,
    "eqangle6_ncoll_cong": match_eqangle6_ncoll_cong,
    "perp_perp_ncoll_para": match_perp_perp_ncoll_para,
    "circle_perp_eqangle": match_circle_perp_eqangle,
    "circle_eqangle_perp": match_circle_eqangle_perp,
    "cyclic_eqangle_cong": match_cyclic_eqangle_cong,
    "midp_perp_cong": match_midp_perp_cong,
    "perp_perp_npara_eqangle": match_perp_perp_npara_eqangle,
    "cyclic_eqangle": match_cyclic_eqangle,
    "eqangle_eqangle_eqangle": match_eqangle_eqangle_eqangle,
    "eqratio_eqratio_eqratio": match_eqratio_eqratio_eqratio,
    "eqratio6_coll_ncoll_eqangle6": match_eqratio6_coll_ncoll_eqangle6,
    "eqangle6_coll_ncoll_eqratio6": match_eqangle6_coll_ncoll_eqratio6,
    "eqangle6_ncoll_cyclic": match_eqangle6_ncoll_cyclic,
}


SKIP_THEOREMS = set()


def set_skip_theorems(theorems: set[str]) -> None:
    SKIP_THEOREMS.update(theorems)


MAX_BRANCH = 50_000


def match_one_theorem(
    g: gh.Graph, cache: Callable[[str], list[tuple[gm.Point, ...]]], theorem: pr.Theorem
) -> Generator[dict[str, gm.Point], None, None]:
    """Match all instances of a single theorem (rule)."""
    if cache is None:
        cache = cache_match(g)

    if theorem.name in SKIP_THEOREMS:
        return []

    if theorem.name.split("_")[-1] in SKIP_THEOREMS:
        return []

    if theorem.name in BUILT_IN_FNS:
        mps = BUILT_IN_FNS[theorem.name](g, cache, theorem)
    else:
        mps = match_generic(g, cache, theorem)

    mappings = []
    for mp in mps:
        mappings.append(mp)
        if len(mappings) > MAX_BRANCH:  # cap branching at this number.
            break

    return mappings


def match_all_theorems(
    g: gh.Graph, theorems: list[pr.Theorem], goal: pr.Clause
) -> dict[pr.Theorem, dict[pr.Theorem, dict[str, gm.Point]]]:
    """Match all instances of all theorems (rules)."""
    cache = cache_match(g)
    # for BFS, collect all potential matches
    # and then do it at the same time
    theorem2mappings = {}

    # Step 1: list all matches
    for _, theorem in theorems.items():
        name = theorem.name
        if name.split("_")[-1] in [
            "acompute",
            "rcompute",
            "fixl",
            "fixc",
            "fixb",
            "fixt",
            "fixp",
        ]:
            if goal and goal.name != name:
                continue

        mappings = match_one_theorem(g, cache, theorem)
        if len(mappings):
            theorem2mappings[theorem] = list(mappings)
    return theorem2mappings


def bfs_one_level(
    g: gh.Graph,
    theorems: list[pr.Theorem],
    level: int,
    controller: pr.Problem,
    verbose: bool = False,
    nm_check: bool = False,
    timeout: int = 600,
) -> tuple[
    list[pr.Dependency],
    dict[str, list[tuple[gm.Point, ...]]],
    dict[str, list[tuple[gm.Point, ...]]],
    int,
]:
    """Forward deduce one breadth-first level."""

    # Step 1: match all theorems:
    theorem2mappings = match_all_theorems(g, theorems, controller.goal)

    # Step 2: traceback for each deduce:
    theorem2deps = {}
    t0 = time.time()
    for theorem, mappings in theorem2mappings.items():
        if time.time() - t0 > timeout:
            break
        mp_deps = []
        for mp in mappings:
            deps = EmptyDependency(level=level, rule_name=theorem.rule_name)
            fail = False  # finding why deps might fail.

            for p in theorem.premise:
                p_args = [mp[a] for a in p.args]
                # Trivial deps.
                if p.name == "cong":
                    a, b, c, d = p_args
                    if {a, b} == {c, d}:
                        continue
                if p.name == "para":
                    a, b, c, d = p_args
                    if {a, b} == {c, d}:
                        continue

                if theorem.name in [
                    "cong_cong_eqangle6_ncoll_contri*",
                    "eqratio6_eqangle6_ncoll_simtri*",
                ]:
                    if p.name in ["eqangle", "eqangle6"]:  # SAS or RAR
                        b, a, b, c, y, x, y, z = p_args
                        if not nm.same_clock(a.num, b.num, c.num, x.num, y.num, z.num):
                            p_args = b, a, b, c, y, z, y, x

                dep = Dependency(p.name, p_args, rule_name="", level=level)
                try:
                    dep = dep.why_me_or_cache(g, level)
                except Exception:
                    fail = True
                    break

                if dep.why is None:
                    fail = True
                    break
                g.cache_dep(p.name, p_args, dep)
                deps.why.append(dep)

            if fail:
                continue

            mp_deps.append((mp, deps))
        theorem2deps[theorem] = mp_deps

    theorem2deps = list(theorem2deps.items())

    # Step 3: add conclusions to graph.
    # Note that we do NOT mix step 2 and 3, strictly going for BFS.
    added = []
    for theorem, mp_deps in theorem2deps:
        for mp, deps in mp_deps:
            if time.time() - t0 > timeout:
                break
            name, args = theorem.conclusion_name_args(mp)
            hash_conclusion = pr.hashed(name, args)
            if hash_conclusion in g.cache:
                continue

            add = g.add_piece(name, args, deps=deps)
            added += add

    branching = len(added)

    # Check if goal is found
    if controller.goal:
        args = []

        for a in controller.goal.args:
            if a in g._name2node:
                a = g._name2node[a]
            elif "/" in a:
                a = create_consts_str(g, a)
            elif a.isdigit():
                a = int(a)
            args.append(a)

        if g.check(controller.goal.name, args):
            return added, {}, {}, branching

    # Run AR, but do NOT apply to the proof state (yet).
    for dep in added:
        g.add_algebra(dep, level)
    derives, eq4s = g.derive_algebra(level, verbose=False)

    branching += sum([len(x) for x in derives.values()])
    branching += sum([len(x) for x in eq4s.values()])

    return added, derives, eq4s, branching


def create_consts_str(g: gh.Graph, s: str) -> gm.Angle | gm.Ratio:
    if "pi/" in s:
        n, d = s.split("pi/")
        n, d = int(n), int(d)
        p0, _ = g.get_or_create_const_ang(n, d)
    else:
        n, d = s.split("/")
        n, d = int(n), int(d)
        p0, _ = g.get_or_create_const_rat(n, d)
    return p0


def do_algebra(g: gh.Graph, added: list[pr.Dependency], verbose: bool = False) -> None:
    for add in added:
        g.add_algebra(add, None)
    derives, eq4s = g.derive_algebra(level=None, verbose=verbose)
    apply_derivations(g, derives)
    apply_derivations(g, eq4s)


def apply_derivations(
    g: gh.Graph, derives: dict[str, list[tuple[gm.Point, ...]]]
) -> list[pr.Dependency]:
    applied = []
    all_derives = list(derives.items())
    for name, args in all_derives:
        for arg in args:
            applied += g.do_algebra(name, arg)
    return applied



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implements DAG-level traceback."""

from typing import Any

import newclid.geometry as gm
import newclid.pretty as pt
import newclid.problem as problem


pretty = pt.pretty


def point_levels(
    setup: list[problem.Dependency], existing_points: list[gm.Point]
) -> list[tuple[set[gm.Point], list[problem.Dependency]]]:
    """Reformat setup into levels of point constructions."""
    levels = []
    for con in setup:
        plevel = max([p.plevel for p in con.args if isinstance(p, gm.Point)])

        while len(levels) - 1 < plevel:
            levels.append((set(), []))

        for p in con.args:
            if not isinstance(p, gm.Point):
                continue
            if existing_points and p in existing_points:
                continue

            levels[p.plevel][0].add(p)

        cons = levels[plevel][1]
        cons.append(con)

    return [(p, c) for p, c in levels if p or c]


def point_log(
    setup: list[problem.Dependency],
    ref_id: dict[tuple[str, ...], int],
    existing_points=list[gm.Point],
) -> list[tuple[list[gm.Point], list[problem.Dependency]]]:
    """Reformat setup into groups of point constructions."""
    log = []

    levels = point_levels(setup, existing_points)

    for points, cons in levels:
        for con in cons:
            if con.hashed() not in ref_id:
                ref_id[con.hashed()] = len(ref_id)

        log.append((points, cons))

    return log


def setup_to_levels(
    setup: list[problem.Dependency],
) -> list[list[problem.Dependency]]:
    """Reformat setup into levels of point constructions."""
    levels = []
    for d in setup:
        plevel = max([p.plevel for p in d.args if isinstance(p, gm.Point)])
        while len(levels) - 1 < plevel:
            levels.append([])

        levels[plevel].append(d)

    levels = [lvl for lvl in levels if lvl]
    return levels


def separate_dependency_difference(
    query: problem.Dependency,
    log: list[tuple[list[problem.Dependency], list[problem.Dependency]]],
) -> tuple[
    list[tuple[list[problem.Dependency], list[problem.Dependency]]],
    list[problem.Dependency],
    list[problem.Dependency],
    set[gm.Point],
    set[gm.Point],
]:
    """Identify and separate the dependency difference."""
    setup = []
    log_, log = log, []
    for prems, cons in log_:
        if not prems:
            setup.extend(cons)
            continue
        cons_ = []
        for con in cons:
            if con.rule_name == "c0":
                setup.append(con)
            else:
                cons_.append(con)
        if not cons_:
            continue

        prems = [p for p in prems if p.name != "ind"]
        log.append((prems, cons_))

    points = set(query.args)
    queue = list(query.args)
    i = 0
    while i < len(queue):
        q = queue[i]
        i += 1
        if not isinstance(q, gm.Point):
            continue
        for p in q.rely_on:
            if p not in points:
                points.add(p)
                queue.append(p)

    setup_, setup, aux_setup, aux_points = setup, [], [], set()
    for con in setup_:
        if con.name == "ind":
            continue
        elif any([p not in points for p in con.args if isinstance(p, gm.Point)]):
            aux_setup.append(con)
            aux_points.update(
                [p for p in con.args if isinstance(p, gm.Point) and p not in points]
            )
        else:
            setup.append(con)

    return log, setup, aux_setup, points, aux_points


def recursive_traceback(
    query: problem.Dependency,
) -> list[tuple[list[problem.Dependency], list[problem.Dependency]]]:
    """Recursively traceback from the query, i.e. the conclusion."""
    visited = set()
    log = []
    stack = []

    def read(q: problem.Dependency) -> None:
        q = q.remove_loop()
        hashed = q.hashed()
        if hashed in visited:
            return

        if hashed[0] in ["ncoll", "npara", "nperp", "diff", "sameside"]:
            return

        nonlocal stack

        stack.append(hashed)
        prems = []

        if q.rule_name != problem.CONSTRUCTION_RULE:
            all_deps = []
            dep_names = set()
            for d in q.why:
                if d.hashed() in dep_names:
                    continue
                dep_names.add(d.hashed())
                all_deps.append(d)

            for d in all_deps:
                h = d.hashed()
                if h not in visited:
                    read(d)
                if h in visited:
                    prems.append(d)

        visited.add(hashed)
        hashs = sorted([d.hashed() for d in prems])
        found = False
        for ps, qs in log:
            if sorted([d.hashed() for d in ps]) == hashs:
                qs += [q]
                found = True
                break
        if not found:
            log.append((prems, [q]))

        stack.pop(-1)

    read(query)

    # post process log: separate multi-conclusion lines
    log_, log = log, []
    for ps, qs in log_:
        for q in qs:
            log.append((ps, [q]))

    return log


def collx_to_coll_setup(
    setup: list[problem.Dependency],
) -> list[problem.Dependency]:
    """Convert collx to coll in setups."""
    result = []
    for level in setup_to_levels(setup):
        hashs = set()
        for dep in level:
            if dep.name == "collx":
                dep.name = "coll"
                dep.args = list(set(dep.args))

            if dep.hashed() in hashs:
                continue
            hashs.add(dep.hashed())
            result.append(dep)

    return result


def collx_to_coll(
    setup: list[problem.Dependency],
    aux_setup: list[problem.Dependency],
    log: list[tuple[list[problem.Dependency], list[problem.Dependency]]],
) -> tuple[
    list[problem.Dependency],
    list[problem.Dependency],
    list[tuple[list[problem.Dependency], list[problem.Dependency]]],
]:
    """Convert collx to coll and dedup."""
    setup = collx_to_coll_setup(setup)
    aux_setup = collx_to_coll_setup(aux_setup)

    con_set = set([p.hashed() for p in setup + aux_setup])
    log_, log = log, []
    for prems, cons in log_:
        prem_set = set()
        prems_, prems = prems, []
        for p in prems_:
            if p.name == "collx":
                p.name = "coll"
                p.args = list(set(p.args))
            if p.hashed() in prem_set:
                continue
            prem_set.add(p.hashed())
            prems.append(p)

        cons_, cons = cons, []
        for c in cons_:
            if c.name == "collx":
                c.name = "coll"
                c.args = list(set(c.args))
            if c.hashed() in con_set:
                continue
            con_set.add(c.hashed())
            cons.append(c)

        if not cons or not prems:
            continue

        log.append((prems, cons))

    return setup, aux_setup, log


def get_logs(
    query: problem.Dependency, g: Any, merge_trivials: bool = False
) -> tuple[
    list[problem.Dependency],
    list[problem.Dependency],
    list[tuple[list[problem.Dependency], list[problem.Dependency]]],
    set[gm.Point],
]:
    """Given a DAG and conclusion N, return the premise, aux, proof."""
    query = query.why_me_or_cache(g, query.level)
    log = recursive_traceback(query)
    log, setup, aux_setup, setup_points, _ = separate_dependency_difference(query, log)

    setup, aux_setup, log = collx_to_coll(setup, aux_setup, log)

    setup, aux_setup, log = shorten_and_shave(setup, aux_setup, log, merge_trivials)

    return setup, aux_setup, log, setup_points


def shorten_and_shave(
    setup: list[problem.Dependency],
    aux_setup: list[problem.Dependency],
    log: list[tuple[list[problem.Dependency], list[problem.Dependency]]],
    merge_trivials: bool = False,
) -> tuple[
    list[problem.Dependency],
    list[problem.Dependency],
    list[tuple[list[problem.Dependency], list[problem.Dependency]]],
]:
    """Shorten the proof by removing unused predicates."""
    log, _ = shorten_proof(log, merge_trivials=merge_trivials)

    all_prems = sum([list(prems) for prems, _ in log], [])
    all_prems = set([p.hashed() for p in all_prems])
    setup = [d for d in setup if d.hashed() in all_prems]
    aux_setup = [d for d in aux_setup if d.hashed() in all_prems]
    return setup, aux_setup, log


def join_prems(
    con: problem.Dependency,
    con2prems: dict[tuple[str, ...], list[problem.Dependency]],
    expanded: set[tuple[str, ...]],
) -> list[problem.Dependency]:
    """Join proof steps with the same premises."""
    h = con.hashed()
    if h in expanded or h not in con2prems:
        return [con]

    result = []
    for p in con2prems[h]:
        result += join_prems(p, con2prems, expanded)
    return result


def shorten_proof(
    log: list[tuple[list[problem.Dependency], list[problem.Dependency]]],
    merge_trivials: bool = False,
) -> tuple[
    list[tuple[list[problem.Dependency], list[problem.Dependency]]],
    dict[tuple[str, ...], list[problem.Dependency]],
]:
    """Join multiple trivials proof steps into one."""
    pops = set()
    con2prem = {}
    for prems, cons in log:
        assert len(cons) == 1
        con = cons[0]
        if con.rule_name == "":
            con2prem[con.hashed()] = prems
        elif not merge_trivials:
            # except for the ones that are premises to non-trivial steps.
            pops.update({p.hashed() for p in prems})

    for p in pops:
        if p in con2prem:
            con2prem.pop(p)

    expanded = set()
    log2 = []
    for i, (prems, cons) in enumerate(log):
        con = cons[0]
        if i < len(log) - 1 and con.hashed() in con2prem:
            continue

        hashs = set()
        new_prems = []

        for p in sum([join_prems(p, con2prem, expanded) for p in prems], []):
            if p.hashed() not in hashs:
                new_prems.append(p)
                hashs.add(p.hashed())

        log2 += [(new_prems, [con])]
        expanded.add(con.hashed())

    return log2, con2prem



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implements the combination DD+AR."""

from __future__ import annotations
import time
import logging

import newclid.dd as dd
import newclid.graph as gh
import newclid.problem as pr
from newclid.problem import Dependency
import newclid.trace_back as trace_back


def saturate_or_goal(
    g: gh.Graph,
    theorems: list[pr.Theorem],
    level_times: list[float],
    p: pr.Problem,
    max_level: int = 100,
    timeout: int = 600,
) -> tuple[
    list[dict[str, list[tuple[gh.Point, ...]]]],
    list[dict[str, list[tuple[gh.Point, ...]]]],
    list[int],
    list[pr.Dependency],
]:
    """Run DD until saturation or goal found."""
    derives = []
    eq4s = []
    branching = []
    all_added = []

    while len(level_times) < max_level:
        level = len(level_times) + 1

        t = time.time()
        added, derv, eq4, n_branching = dd.bfs_one_level(
            g, theorems, level, p, verbose=False, nm_check=True, timeout=timeout
        )
        all_added += added
        branching.append(n_branching)

        derives.append(derv)
        eq4s.append(eq4)
        level_time = time.time() - t

        logging.info(f"Depth {level}/{max_level} time = {level_time}")
        level_times.append(level_time)

        if p.goal is not None:
            goal_args = list(map(lambda x: g.get(x, lambda: int(x)), p.goal.args))
            if g.check(p.goal.name, goal_args):  # found goal
                break

        if not added:  # saturated
            break

        if level_time > timeout:
            break

    return derives, eq4s, branching, all_added


def solve(
    g: gh.Graph,
    theorems: list[pr.Problem],
    controller: pr.Problem,
    max_level: int = 1000,
    timeout: int = 600,
) -> tuple[gh.Graph, list[float], str, list[int], list[pr.Dependency]]:
    """Alternate between DD and AR until goal is found."""
    status = "saturated"
    level_times = []

    dervs, eq4 = g.derive_algebra(level=0, verbose=False)
    derives = [dervs]
    eq4s = [eq4]
    branches = []
    all_added = []

    while len(level_times) < max_level:
        dervs, eq4, next_branches, added = saturate_or_goal(
            g, theorems, level_times, controller, max_level, timeout=timeout
        )
        all_added += added

        derives += dervs
        eq4s += eq4
        branches += next_branches

        # Now, it is either goal or saturated
        if controller.goal is not None:
            goal_args = g.names2points(controller.goal.args)
            if g.check(controller.goal.name, goal_args):  # found goal
                status = "solved"
                break

        if not derives:  # officially saturated.
            break

        # Now we resort to algebra derivations.
        added = []
        while derives and not added:
            added += dd.apply_derivations(g, derives.pop(0))

        if added:
            continue

        # Final help from AR.
        while eq4s and not added:
            added += dd.apply_derivations(g, eq4s.pop(0))

        all_added += added

        if not added:  # Nothing left. saturated.
            break

    return g, level_times, status, branches, all_added


def get_proof_steps(
    g: gh.Graph, goal: pr.Clause, merge_trivials: bool = False
) -> tuple[
    list[pr.Dependency],
    list[pr.Dependency],
    list[tuple[list[pr.Dependency], list[pr.Dependency]]],
    dict[tuple[str, ...], int],
]:
    """Extract proof steps from the built DAG."""
    goal_args = g.names2nodes(goal.args)
    query = Dependency(goal.name, goal_args, None, None)

    setup, aux, log, setup_points = trace_back.get_logs(
        query, g, merge_trivials=merge_trivials
    )

    refs = {}
    setup = trace_back.point_log(setup, refs, set())
    aux = trace_back.point_log(aux, refs, setup_points)

    setup = [(prems, [tuple(p)]) for p, prems in setup]
    aux = [(prems, [tuple(p)]) for p, prems in aux]

    return setup, aux, log, refs



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Unit testing for the trace_back code."""

import pytest
import pytest_check as check

from newclid.api import GeometricSolverBuilder
import newclid.problem as pr
import newclid.trace_back as tb


class TestTraceback:
    @pytest.fixture(autouse=True)
    def setUpClass(self):
        self.solver_builder = GeometricSolverBuilder()

    def test_orthocenter_dependency_difference(self):
        solver = self.solver_builder.load_problem_from_txt(
            "a b c = triangle a b c; "
            "d = on_tline d b a c, on_tline d c a b; "
            "e = on_line e a c, on_line e b d "
            "? perp a d b c"
        ).build()

        solver.run()

        goal_args = solver.proof_state.names2nodes(solver.goal.args)
        query = pr.Dependency(solver.goal.name, goal_args, None, None)
        setup, aux, _, _ = tb.get_logs(query, solver.proof_state, merge_trivials=False)

        # Convert each predicates to its hash string:
        setup = [p.hashed() for p in setup]
        aux = [p.hashed() for p in aux]

        check.equal(
            set(setup), {("perp", "a", "c", "b", "d"), ("perp", "a", "b", "c", "d")}
        )

        check.equal(set(aux), {("coll", "a", "c", "e"), ("coll", "b", "d", "e")})



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Unit tests for geometry.py."""

import pytest_check as check

import newclid.geometry as gm


class TestGeometry:
    def _setup_equality_example(self):
        # Create 4 nodes a, b, c, d
        # and their lengths
        a = gm.Segment("a")
        la = gm.Length("l(a)")
        a.connect_to(la)
        la.connect_to(a)

        b = gm.Segment("b")
        lb = gm.Length("l(b)")
        b.connect_to(lb)
        lb.connect_to(b)

        c = gm.Segment("c")
        lc = gm.Length("l(c)")
        c.connect_to(lc)
        lc.connect_to(c)

        d = gm.Segment("d")
        ld = gm.Length("l(d)")
        d.connect_to(ld)
        ld.connect_to(d)

        # Now let a=b, b=c, a=c, c=d
        la.merge([lb], "fact1")
        lb.merge([lc], "fact2")
        la.merge([lc], "fact3")
        lc.merge([ld], "fact4")
        return a, b, c, d, la, lb, lc, ld

    def test_merged_node_representative(self):
        _, _, _, _, la, lb, lc, ld = self._setup_equality_example()

        # all nodes are now represented by la.
        check.equal(la.rep(), la)
        check.equal(lb.rep(), la)
        check.equal(lc.rep(), la)
        check.equal(ld.rep(), la)

    def test_merged_node_equivalence(self):
        _, _, _, _, la, lb, lc, ld = self._setup_equality_example()
        # all la, lb, lc, ld are equivalent
        check.equal(set(la.equivs()), {la, lb, lc, ld})
        check.equal(set(lb.equivs()), {la, lb, lc, ld})
        check.equal(set(lc.equivs()), {la, lb, lc, ld})
        check.equal(set(ld.equivs()), {la, lb, lc, ld})

    def test_bfs_for_equality_transitivity(self):
        a, _, _, d, _, _, _, _ = self._setup_equality_example()

        # check that a==d because fact3 & fact4, not fact1 & fact2
        check.equal(set(gm.why_equal(a, d)), {"fact3", "fact4"})



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Unit tests for graph.py."""

import pytest
import pytest_check as check

from newclid.api import GeometricSolverBuilder
import newclid.numericals as nm


MAX_LEVEL = 10


class TestGraph:
    @pytest.fixture(autouse=True)
    def setUpClass(self):
        self.solver = (
            GeometricSolverBuilder()
            .load_problem_from_txt(
                "a b c = triangle a b c; "
                "h = orthocenter a b c; "
                "h1 = foot a b c; "
                "h2 = foot b c a; "
                "h3 = foot c a b; "
                "g1 g2 g3 g = centroid g1 g2 g3 g a b c; "
                "o = circle a b c "
                "? coll h g o",
                translate=False,
            )
            .build()
        )
        self.graph = self.solver.proof_state

    def test_build_graph_points(self):
        g = self.graph

        all_points = g.all_points()
        check.equal(
            {p.name for p in all_points},
            {"a", "b", "c", "g", "h", "o", "g1", "g2", "g3", "h1", "h2", "h3"},
        )

    def test_build_graph_predicates(self):
        gr = self.graph

        a, b, c, g, h, o, g1, g2, g3, h1, h2, h3 = gr.names2points(
            ["a", "b", "c", "g", "h", "o", "g1", "g2", "g3", "h1", "h2", "h3"]
        )

        # Explicit statements:
        check.is_true(gr.check_cong([b, g1, g1, c]))
        check.is_true(gr.check_cong([c, g2, g2, a]))
        check.is_true(gr.check_cong([a, g3, g3, b]))
        check.is_true(gr.check_perp([a, h1, b, c]))
        check.is_true(gr.check_perp([b, h2, c, a]))
        check.is_true(gr.check_perp([c, h3, a, b]))
        check.is_true(gr.check_cong([o, a, o, b]))
        check.is_true(gr.check_cong([o, b, o, c]))
        check.is_true(gr.check_cong([o, a, o, c]))
        check.is_true(gr.check_coll([a, g, g1]))
        check.is_true(gr.check_coll([b, g, g2]))
        check.is_true(gr.check_coll([g1, b, c]))
        check.is_true(gr.check_coll([g2, c, a]))
        check.is_true(gr.check_coll([g3, a, b]))
        check.is_true(gr.check_perp([a, h, b, c]))
        check.is_true(gr.check_perp([b, h, c, a]))

        # These are NOT part of the premises:
        check.is_false(gr.check_perp([c, h, a, b]))
        check.is_false(gr.check_coll([c, g, g3]))

        # These are automatically inferred by the graph datastructure:
        check.is_true(gr.check_eqangle([a, h1, b, c, b, h2, c, a]))
        check.is_true(gr.check_eqangle([a, h1, b, h2, b, c, c, a]))
        check.is_true(gr.check_eqratio([b, g1, g1, c, c, g2, g2, a]))
        check.is_true(gr.check_eqratio([b, g1, g1, c, o, a, o, b]))
        check.is_true(gr.check_para([a, h, a, h1]))
        check.is_true(gr.check_para([b, h, b, h2]))
        check.is_true(gr.check_coll([a, h, h1]))
        check.is_true(gr.check_coll([b, h, h2]))

    def test_enumerate_colls(self):
        g = self.graph

        for a, b, c in g.all_colls():
            check.is_true(g.check_coll([a, b, c]))
            check.is_true(nm.check_coll([a.num, b.num, c.num]))

    def test_enumerate_paras(self):
        g = self.graph

        for a, b, c, d in g.all_paras():
            check.is_true(g.check_para([a, b, c, d]))
            check.is_true(nm.check_para([a.num, b.num, c.num, d.num]))

    def test_enumerate_perps(self):
        g = self.graph

        for a, b, c, d in g.all_perps():
            check.is_true(g.check_perp([a, b, c, d]))
            check.is_true(nm.check_perp([a.num, b.num, c.num, d.num]))

    def test_enumerate_congs(self):
        g = self.graph

        for a, b, c, d in g.all_congs():
            check.is_true(g.check_cong([a, b, c, d]))
            check.is_true(nm.check_cong([a.num, b.num, c.num, d.num]))

    def test_enumerate_eqangles(self):
        g = self.graph

        for a, b, c, d, x, y, z, t in g.all_eqangles_8points():
            check.is_true(g.check_eqangle([a, b, c, d, x, y, z, t]))
            check.is_true(
                nm.check_eqangle(
                    [a.num, b.num, c.num, d.num, x.num, y.num, z.num, t.num]
                )
            )

    def test_enumerate_eqratios(self):
        g = self.graph

        for a, b, c, d, x, y, z, t in g.all_eqratios_8points():
            check.is_true(g.check_eqratio([a, b, c, d, x, y, z, t]))
            check.is_true(
                nm.check_eqratio(
                    [a.num, b.num, c.num, d.num, x.num, y.num, z.num, t.num]
                )
            )

    def test_enumerate_cyclics(self):
        g = self.graph

        for a, b, c, d, x, y, z, t in g.all_cyclics():
            check.is_true(g.check_cyclic([a, b, c, d, x, y, z, t]))
            check.is_true(nm.check_cyclic([a.num, b.num, c.num, d.num]))

    def test_enumerate_midps(self):
        g = self.graph

        for a, b, c in g.all_midps():
            check.is_true(g.check_midp([a, b, c]))
            check.is_true(nm.check_midp([a.num, b.num, c.num]))

    def test_enumerate_circles(self):
        g = self.graph

        for a, b, c, d in g.all_circles():
            check.is_true(g.check_circle([a, b, c, d]))
            check.is_true(nm.check_circle([a.num, b.num, c.num, d.num]))



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implements geometric objects used in the graph representation."""
from __future__ import annotations
from collections import defaultdict
from typing import Any, Type


class Node:
    r"""Node in the proof state graph.

  Can be Point, Line, Circle, etc.

  Each node maintains a merge history to
  other nodes if they are (found out to be) equivalent

    a -> b -
            \
         c -> d -> e -> f -> g

  d.merged_to = e
  d.rep = g
  d.merged_from = {a, b, c, d}
  d.equivs = {a, b, c, d, e, f, g}
  """

    def __init__(self, name: str = "", graph: Any = None):
        self.name = name or str(self)
        self.graph = graph

        self.edge_graph = {}
        # Edge graph: what other nodes is connected to this node.
        # edge graph = {
        #   other1: {self1: deps, self2: deps},
        #   other2: {self2: deps, self3: deps}
        # }

        self.merge_graph = {}
        # Merge graph: history of merges with other nodes.
        # merge_graph = {self1: {self2: deps1, self3: deps2}}

        self.rep_by = None  # represented by.
        self.members = {self}

        self._val = None
        self._obj = None

        self.deps = []

        # numerical representation.
        self.num = None
        self.change = set()  # what other nodes' num rely on this node?

    def set_rep(self, node: Node) -> None:
        if node == self:
            return
        self.rep_by = node
        node.merge_edge_graph(self.edge_graph)
        node.members.update(self.members)

    def rep(self) -> Node:
        x = self
        while x.rep_by:
            x = x.rep_by
        return x

    def why_rep(self) -> list[Any]:
        return self.why_equal([self.rep()], None)

    def rep_and_why(self) -> tuple[Node, list[Any]]:
        rep = self.rep()
        return rep, self.why_equal([rep], None)

    def neighbors(
        self, oftype: Type[Node], return_set: bool = False, do_rep: bool = True
    ) -> list[Node]:
        """Neighbors of this node in the proof state graph."""
        if do_rep:
            rep = self.rep()
        else:
            rep = self
        result = set()

        for n in rep.edge_graph:
            if oftype is None or oftype and isinstance(n, oftype):
                if do_rep:
                    result.add(n.rep())
                else:
                    result.add(n)

        if return_set:
            return result
        return list(result)

    def merge_edge_graph(
        self, new_edge_graph: dict[Node, dict[Node, list[Node]]]
    ) -> None:
        for x, xdict in new_edge_graph.items():
            if x in self.edge_graph:
                self.edge_graph[x].update(dict(xdict))
            else:
                self.edge_graph[x] = dict(xdict)

    def merge(self, nodes: list[Node], deps: list[Any]) -> None:
        for node in nodes:
            self.merge_one(node, deps)

    def merge_one(self, node: Node, deps: list[Any]) -> None:
        node.rep().set_rep(self.rep())

        if node in self.merge_graph:
            return

        self.merge_graph[node] = deps
        node.merge_graph[self] = deps

    def is_val(self, node: Node) -> bool:
        return (
            isinstance(self, Line)
            and isinstance(node, Direction)
            or isinstance(self, Segment)
            and isinstance(node, Length)
            or isinstance(self, Angle)
            and isinstance(node, Measure)
            or isinstance(self, Ratio)
            and isinstance(node, Value)
        )

    def set_val(self, node: Node) -> None:
        self._val = node

    def set_obj(self, node: Node) -> None:
        self._obj = node

    @property
    def val(self) -> Node:
        if self._val is None:
            return None
        return self._val.rep()

    @property
    def obj(self) -> Node:
        if self._obj is None:
            return None
        return self._obj.rep()

    def equivs(self) -> set[Node]:
        return self.rep().members

    def connect_to(self, node: Node, deps: list[Any] = None) -> None:
        rep = self.rep()

        if node in rep.edge_graph:
            rep.edge_graph[node].update({self: deps})
        else:
            rep.edge_graph[node] = {self: deps}

        if self.is_val(node):
            self.set_val(node)
            node.set_obj(self)

    def equivs_upto(self, level: int) -> dict[Node, Node]:
        """What are the equivalent nodes up to a certain level."""
        parent = {self: None}
        visited = set()
        queue = [self]
        i = 0

        while i < len(queue):
            current = queue[i]
            i += 1
            visited.add(current)

            for neighbor in current.merge_graph:
                if (
                    level is not None
                    and current.merge_graph[neighbor].level is not None
                    and current.merge_graph[neighbor].level >= level
                ):
                    continue
                if neighbor not in visited:
                    queue.append(neighbor)
                    parent[neighbor] = current

        return parent

    def why_equal(self, others: list[Node], level: int) -> list[Any]:
        """BFS why this node is equal to other nodes."""
        others = set(others)
        found = 0

        parent = {}
        queue = [self]
        i = 0

        while i < len(queue):
            current = queue[i]
            if current in others:
                found += 1
            if found == len(others):
                break

            i += 1

            for neighbor in current.merge_graph:
                if (
                    level is not None
                    and current.merge_graph[neighbor].level is not None
                    and current.merge_graph[neighbor].level >= level
                ):
                    continue
                if neighbor not in parent:
                    queue.append(neighbor)
                    parent[neighbor] = current

        return bfs_backtrack(self, others, parent)

    def why_equal_groups(
        self, groups: list[list[Node]], level: int
    ) -> tuple[list[Any], list[Node]]:
        """BFS for why self is equal to at least one member of each group."""
        others = [None for _ in groups]
        found = 0

        parent = {}
        queue = [self]
        i = 0

        while i < len(queue):
            current = queue[i]

            for j, grp in enumerate(groups):
                if others[j] is None and current in grp:
                    others[j] = current
                    found += 1

            if found == len(others):
                break

            i += 1

            for neighbor in current.merge_graph:
                if (
                    level is not None
                    and current.merge_graph[neighbor].level is not None
                    and current.merge_graph[neighbor].level >= level
                ):
                    continue
                if neighbor not in parent:
                    queue.append(neighbor)
                    parent[neighbor] = current

        return bfs_backtrack(self, others, parent), others

    def why_val(self, level: int) -> list[Any]:
        return self._val.why_equal([self.val], level)

    def why_connect(self, node: Node, level: int = None) -> list[Any]:
        rep = self.rep()
        equivs = list(rep.edge_graph[node].keys())
        if not equivs:
            return None
        equiv = equivs[0]
        dep = rep.edge_graph[node][equiv]
        return [dep] + self.why_equal(equiv, level)


def why_connect(*pairs: list[tuple[Node, Node]]) -> list[Any]:
    result = []
    for node1, node2 in pairs:
        result += node1.why_connect(node2)
    return result


def is_equiv(x: Node, y: Node, level: int = None) -> bool:
    level = level or float("inf")
    return x.why_equal([y], level) is not None


def is_equal(x: Node, y: Node, level: int = None) -> bool:
    if x == y:
        return True
    if x._val is None or y._val is None:
        return False
    if x.val != y.val:
        return False
    return is_equiv(x._val, y._val, level)


def bfs_backtrack(root: Node, leafs: list[Node], parent: dict[Node, Node]) -> list[Any]:
    """Return the path given BFS trace of parent nodes."""
    backtracked = {root}  # no need to backtrack further when touching this set.
    deps = []
    for node in leafs:
        if node is None:
            return None
        if node in backtracked:
            continue
        if node not in parent:
            return None
        while node not in backtracked:
            backtracked.add(node)
            deps.append(node.merge_graph[parent[node]])
            node = parent[node]

    return deps


class Point(Node):
    pass


class Line(Node):
    """Node of type Line."""

    def new_val(self) -> Direction:
        return Direction()

    def why_coll(self, points: list[Point], level: int = None) -> list[Any]:
        """Why points are connected to self."""
        level = level or float("inf")

        groups = []
        for p in points:
            group = [
                level
                for level, d in self.edge_graph[p].items()
                if d is None or d.level < level
            ]
            if not group:
                return None
            groups.append(group)

        min_deps = None
        for line in groups[0]:
            deps, others = line.why_equal_groups(groups[1:], level)
            if deps is None:
                continue
            for p, o in zip(points, [line] + others):
                deps.append(self.edge_graph[p][o])
            if min_deps is None or len(deps) < len(min_deps):
                min_deps = deps

        if min_deps is None:
            return None
        return [d for d in min_deps if d is not None]


class Segment(Node):
    def new_val(self) -> Length:
        return Length()


class Circle(Node):
    """Node of type Circle."""

    def why_cyclic(self, points: list[Point], level: int = None) -> list[Any]:
        """Why points are connected to self."""
        level = level or float("inf")

        groups = []
        for p in points:
            group = [
                c for c, d in self.edge_graph[p].items() if d is None or d.level < level
            ]
            if not group:
                return None
            groups.append(group)

        min_deps = None
        for circle in groups[0]:
            deps, others = circle.why_equal_groups(groups[1:], level)
            if deps is None:
                continue
            for p, o in zip(points, [circle] + others):
                deps.append(self.edge_graph[p][o])

            if min_deps is None or len(deps) < len(min_deps):
                min_deps = deps

        if min_deps is None:
            return None
        return [d for d in min_deps if d is not None]


def why_equal(x: Node, y: Node, level: int = None) -> list[Any]:
    if x == y:
        return []
    if not x._val or not y._val:
        return None
    if x._val == y._val:
        return []
    return x._val.why_equal([y._val], level)


class Direction(Node):
    pass


def get_lines_thru_all(*points: Point) -> list[Line]:
    line2count = defaultdict(lambda: 0)
    points = set(points)
    for p in points:
        for line_neighbor in p.neighbors(Line):
            line2count[line_neighbor] += 1
    return [line for line, count in line2count.items() if count == len(points)]


def line_of_and_why(
    points: list[Point], level: int | None = None
) -> tuple[Line, list[Any]]:
    """Why points are collinear."""
    for l0 in get_lines_thru_all(*points):
        for line in l0.equivs():
            if all([p in line.edge_graph for p in points]):
                x, y = line.points
                colls = list({x, y} | set(points))
                # if len(colls) < 3:
                #   return l, []
                why = line.why_coll(colls, level)
                if why is not None:
                    return line, why

    return None, None


def get_circles_thru_all(*points: list[Point]) -> list[Circle]:
    circle2count = defaultdict(lambda: 0)
    points = set(points)
    for p in points:
        for c in p.neighbors(Circle):
            circle2count[c] += 1
    return [c for c, count in circle2count.items() if count == len(points)]


def circle_of_and_why(
    points: list[Point], level: int = None
) -> tuple[Circle, list[Any]]:
    """Why points are concyclic."""
    for c0 in get_circles_thru_all(*points):
        for c in c0.equivs():
            if all([p in c.edge_graph for p in points]):
                cycls = list(set(points))
                why = c.why_cyclic(cycls, level)
                if why is not None:
                    return c, why

    return None, None


def name_map(struct: Any) -> Any:
    if isinstance(struct, list):
        return [name_map(x) for x in struct]
    elif isinstance(struct, tuple):
        return tuple([name_map(x) for x in struct])
    elif isinstance(struct, set):
        return set([name_map(x) for x in struct])
    elif isinstance(struct, dict):
        return {name_map(x): name_map(y) for x, y in struct.items()}
    else:
        return getattr(struct, "name", "")


class Angle(Node):
    """Node of type Angle."""

    def new_val(self) -> Measure:
        return Measure()

    def set_directions(self, d1: Direction, d2: Direction) -> None:
        self._d = d1, d2

    @property
    def directions(self) -> tuple[Direction, Direction]:
        d1, d2 = self._d
        if d1 is None or d2 is None:
            return d1, d2
        return d1.rep(), d2.rep()


class Measure(Node):
    pass


class Length(Node):
    pass


class Ratio(Node):
    """Node of type Ratio."""

    def new_val(self) -> Value:
        return Value()

    def set_lengths(self, l1: Length, l2: Length) -> None:
        self._l = l1, l2

    @property
    def lengths(self) -> tuple[Length, Length]:
        l1, l2 = self._l
        if l1 is None or l2 is None:
            return l1, l2
        return l1.rep(), l2.rep()


class Value(Node):
    pass


def all_angles(
    d1: Direction, d2: Direction, level: int = None
) -> tuple[Angle, list[Direction], list[Direction]]:
    level = level or float("inf")
    d1s = d1.equivs_upto(level)
    d2s = d2.equivs_upto(level)

    for ang in d1.rep().neighbors(Angle):
        d1_, d2_ = ang._d
        if d1_ in d1s and d2_ in d2s:
            yield ang, d1s, d2s


def all_ratios(d1, d2, level=None) -> tuple[Angle, list[Direction], list[Direction]]:
    level = level or float("inf")
    d1s = d1.equivs_upto(level)
    d2s = d2.equivs_upto(level)

    for ang in d1.rep().neighbors(Ratio):
        d1_, d2_ = ang._l
        if d1_ in d1s and d2_ in d2s:
            yield ang, d1s, d2s


RANKING = {
    Point: 0,
    Line: 1,
    Segment: 2,
    Circle: 3,
    Direction: 4,
    Length: 5,
    Angle: 6,
    Ratio: 7,
    Measure: 8,
    Value: 9,
}


def val_type(x: Node) -> Type[Node]:
    if isinstance(x, Line):
        return Direction
    if isinstance(x, Segment):
        return Length
    if isinstance(x, Angle):
        return Measure
    if isinstance(x, Ratio):
        return Value



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implementing Algebraic Reasoning (AR)."""

from collections import defaultdict
from fractions import Fraction as frac
from typing import Any, Generator

import newclid.geometry as gm
import numpy as np
import newclid.problem as pr
from scipy import optimize


class InfQuotientError(Exception):
    pass


def _gcd(x: int, y: int) -> int:
    while y:
        x, y = y, x % y
    return x


def simplify(n: int, d: int) -> tuple[int, int]:
    g = _gcd(n, d)
    return (n // g, d // g)


# maximum denominator for a fraction.
MAX_DENOMINATOR = 1000000

# tolerance for fraction approximation
TOL = 1e-15


def get_quotient(v: float) -> tuple[int, int]:
    n = v
    d = 1
    while abs(n - round(n)) > TOL:
        d += 1
        n += v
        if d > MAX_DENOMINATOR:
            e = InfQuotientError(v)
            raise e

    n = int(round(n))
    return simplify(n, d)


def fix_v(v: float) -> float:
    n, d = get_quotient(v)
    return n / d


def fix(e: dict[str, float]) -> dict[str, float]:
    return {k: fix_v(v) for k, v in e.items()}


def frac_string(f: frac) -> str:
    n, d = get_quotient(f)
    return f"{n}/{d}"


def hashed(e: dict[str, float]) -> tuple[tuple[str, float], ...]:
    return tuple(sorted(list(e.items())))


def is_zero(e: dict[str, float]) -> bool:
    return len(strip(e)) == 0


def strip(e: dict[str, float]) -> dict[str, float]:
    return {v: c for v, c in e.items() if c != 0}


def plus(e1: dict[str, float], e2: dict[str, float]) -> dict[str, float]:
    e = dict(e1)
    for v, c in e2.items():
        if v in e:
            e[v] += c
        else:
            e[v] = c
    return strip(e)


def plus_all(*es: list[dict[str, float]]) -> dict[str, float]:
    result = {}
    for e in es:
        result = plus(result, e)
    return result


def mult(e: dict[str, float], m: float) -> dict[str, float]:
    return {v: m * c for v, c in e.items()}


def minus(e1: dict[str, float], e2: dict[str, float]) -> dict[str, float]:
    return plus(e1, mult(e2, -1))


def div(e1: dict[str, float], e2: dict[str, float]) -> float:
    """Divide e1 by e2."""
    e1 = strip(e1)
    e2 = strip(e2)
    if set(e1.keys()) != set(e2.keys()):
        return None

    n, d = None, None

    for v, c1 in e1.items():
        c2 = e2[v]  # we want c1/c2 = n/d => c1*d=c2*n
        if n is not None and c1 * d != c2 * n:
            return None
        n, d = c1, c2
    return frac(n) / frac(d)


def recon(e: dict[str, float], const: str) -> tuple[str, dict[str, float]]:
    """Reconcile one variable in the expression e=0, given const."""
    e = strip(e)
    if len(e) == 0:
        return None

    v0 = None
    for v in e:
        if v != const:
            v0 = v
            break
    if v0 is None:
        return v0

    c0 = e.pop(v0)
    return v0, {v: -c / c0 for v, c in e.items()}


def replace(e: dict[str, float], v0: str, e0: dict[str, float]) -> dict[str, float]:
    if v0 not in e:
        return e
    e = dict(e)
    m = e.pop(v0)
    return plus(e, mult(e0, m))


def comb2(elems: list[Any]) -> Generator[tuple[Any, Any], None, None]:
    if len(elems) < 1:
        return
    for i, e1 in enumerate(elems[:-1]):
        for e2 in elems[i + 1 :]:
            yield e1, e2


def perm2(elems: list[Any]) -> Generator[tuple[Any, Any], None, None]:
    for e1, e2 in comb2(elems):
        yield e1, e2
        yield e2, e1


def chain2(elems: list[Any]) -> Generator[tuple[Any, Any], None, None]:
    if len(elems) < 2:
        return
    for i, e1 in enumerate(elems[:-1]):
        yield e1, elems[i + 1]


def update_groups(
    groups1: list[Any], groups2: list[Any]
) -> tuple[list[Any], list[tuple[Any, Any]], list[list[Any]]]:
    """Update groups of equivalent elements.

    Given groups1 = [set1, set2, set3, ..]
    where all elems within each set_i is defined to be "equivalent" to each other.
    (but not across the sets)

    Incoming groups2 = [set1, set2, ...] similar to set1 - it is the
    additional equivalent information on elements in groups1.

    Return the new updated groups1 and the set of links
    that make it that way.

    Example:
      groups1 = [{1, 2}, {3, 4, 5}, {6, 7}]
      groups2 = [{2, 3, 8}, {9, 10, 11}]

    => new groups1 and links:
      groups1 = [{1, 2, 3, 4, 5, 8}, {6, 7}, {9, 10, 11}]
      links = (2, 3), (3, 8), (9, 10), (10, 11)

    Explain: since groups2 says 2 and 3 are equivalent (with {2, 3, 8}),
    then {1, 2} and {3, 4, 5} in groups1 will be merged,
    because 2 and 3 each belong to those 2 groups.
    Additionally 8 also belong to this same group.
    {3, 4, 5} is left alone, while {9, 10, 11} is a completely new set.

    The links to make this all happens is:
    (2, 3): to merge {1, 2} and {3, 4, 5}
    (3, 8): to link 8 into the merged({1, 2, 3, 4, 5})
    (9, 10) and (10, 11): to make the new group {9, 10, 11}

    Args:
      groups1: a list of sets.
      groups2: a list of sets.

    Returns:
      groups1, links, history: result of the update.
    """
    history = []
    links = []
    for g2 in groups2:
        joins = [None] * len(groups1)  # mark which one in groups1 is merged
        merged_g1 = set()  # merge them into this.
        old = None  # any elem in g2 that belong to any set in groups1 (old)
        new = []  # all elem in g2 that is new

        for e in g2:
            found = False
            for i, g1 in enumerate(groups1):
                if e not in g1:
                    continue

                found = True
                if joins[i]:
                    continue

                joins[i] = True
                merged_g1.update(g1)

                if old is not None:
                    links.append((old, e))  # link to make merging happen.
                old = e

            if not found:  # e is new!
                new.append(e)

        # now chain elems in new together.
        if old is not None and new:
            links.append((old, new[0]))
            merged_g1.update(new)

        links += chain2(new)

        new_groups1 = []
        if merged_g1:  # put the merged_g1 in first
            new_groups1.append(merged_g1)

        # put the remaining (unjoined) groups in
        new_groups1 += [g1 for j, g1 in zip(joins, groups1) if not j]

        if old is None and new:
            new_groups1 += [set(new)]

        groups1 = new_groups1
        history.append(groups1)

    return groups1, links, history


class Table:
    """The coefficient matrix."""

    def __init__(self, const: str = "1"):
        self.const = const
        self.v2e = {}
        self.add_free(const)  # the table {var: expression}

        # to cache what is already derived/inputted
        self.eqs = set()
        self.groups = []  # groups of equal pairs.

        # for why (linprog)
        self.c = []
        self.v2i = {}  # v -> index of row in A.
        self.deps = []  # equal number of columns.
        self.A = np.zeros([0, 0])
        self.do_why = True

    def add_free(self, v: str) -> None:
        self.v2e[v] = {v: frac(1)}

    def replace(self, v0: str, e0: dict[str, float]) -> None:
        for v, e in list(self.v2e.items()):
            self.v2e[v] = replace(e, v0, e0)

    def add_expr(self, vc: list[tuple[str, float]]) -> bool:
        """Add a new equality, represented by the list of tuples vc=[(v, c), ..]."""
        result = {}
        free = []

        for v, c in vc:
            c = frac(c)
            if v in self.v2e:
                result = plus(result, mult(self.v2e[v], c))
            else:
                free += [(v, c)]

        if free == []:
            if is_zero(self.modulo(result)):
                return False
            result = recon(result, self.const)
            if result is None:
                return False
            v, e = result
            self.replace(v, e)

        elif len(free) == 1:
            v, m = free[0]
            self.v2e[v] = mult(result, frac(-1, m))

        else:
            dependent_v = None
            for v, m in free:
                if dependent_v is None and v != self.const:
                    dependent_v = (v, m)
                    continue

                self.add_free(v)
                result = plus(result, {v: m})

            v, m = dependent_v
            self.v2e[v] = mult(result, frac(-1, m))

        return True

    def register(self, vc: list[tuple[str, float]], dep: pr.Dependency) -> None:
        """Register a new equality vc=[(v, c), ..] with traceback dependency dep."""
        result = plus_all(*[{v: c} for v, c in vc])
        if is_zero(result):
            return

        vs, _ = zip(*vc)
        for v in vs:
            if v not in self.v2i:
                self.v2i[v] = len(self.v2i)

        (m, n), lenght = self.A.shape, len(self.v2i)
        if lenght > m:
            self.A = np.concatenate([self.A, np.zeros([lenght - m, n])], 0)

        new_column = np.zeros([len(self.v2i), 2])  # N, 2
        for v, c in vc:
            new_column[self.v2i[v], 0] += float(c)
            new_column[self.v2i[v], 1] -= float(c)

        self.A = np.concatenate([self.A, new_column], 1)
        self.c += [1.0, -1.0]
        self.deps += [dep]

    def register2(self, a: str, b: str, m: float, n: float, dep: pr.Dependency) -> None:
        self.register([(a, m), (b, -n)], dep)

    def register3(self, a: str, b: str, f: float, dep: pr.Dependency) -> None:
        self.register([(a, 1), (b, -1), (self.const, -f)], dep)

    def register4(self, a: str, b: str, c: str, d: str, dep: pr.Dependency) -> None:
        self.register([(a, 1), (b, -1), (c, -1), (d, 1)], dep)

    def why(self, e: dict[str, float]) -> list[Any]:
        """AR traceback == MILP."""
        if not self.do_why:
            return []
        # why expr == 0?
        # Solve min(c^Tx) s.t. A_eq * x = b_eq, x >= 0
        e = strip(e)
        if not e:
            return []

        b_eq = [0] * len(self.v2i)
        for v, c in e.items():
            b_eq[self.v2i[v]] += float(c)

        try:
            x = optimize.linprog(c=self.c, A_eq=self.A, b_eq=b_eq, method="highs")["x"]
        except ValueError:
            x = optimize.linprog(
                c=self.c,
                A_eq=self.A,
                b_eq=b_eq,
            )["x"]

        deps = []
        for i, dep in enumerate(self.deps):
            if x[2 * i] > 1e-12 or x[2 * i + 1] > 1e-12:
                if dep not in deps:
                    deps.append(dep)
        return deps

    def record_eq(self, v1: str, v2: str, v3: str, v4: str) -> None:
        self.eqs.add((v1, v2, v3, v4))
        self.eqs.add((v2, v1, v4, v3))
        self.eqs.add((v3, v4, v1, v2))
        self.eqs.add((v4, v3, v2, v1))

    def check_record_eq(self, v1: str, v2: str, v3: str, v4: str) -> bool:
        if (v1, v2, v3, v4) in self.eqs:
            return True
        if (v2, v1, v4, v3) in self.eqs:
            return True
        if (v3, v4, v1, v2) in self.eqs:
            return True
        if (v4, v3, v2, v1) in self.eqs:
            return True
        return False

    def add_eq2(self, a: str, b: str, m: float, n: float, dep: pr.Dependency) -> None:
        # a/b = m/n
        if not self.add_expr([(a, m), (b, -n)]):
            return []
        self.register2(a, b, m, n, dep)

    def add_eq3(self, a: str, b: str, f: float, dep: pr.Dependency) -> None:
        # a - b = f * constant
        self.eqs.add((a, b, frac(f)))
        self.eqs.add((b, a, frac(1 - f)))

        if not self.add_expr([(a, 1), (b, -1), (self.const, -f)]):
            return []

        self.register3(a, b, f, dep)

    def add_eq4(self, a: str, b: str, c: str, d: str, dep: pr.Dependency) -> None:
        # a - b = c - d
        self.record_eq(a, b, c, d)
        self.record_eq(a, c, b, d)

        expr = list(minus({a: 1, b: -1}, {c: 1, d: -1}).items())

        if not self.add_expr(expr):
            return []

        self.register4(a, b, c, d, dep)
        self.groups, _, _ = update_groups(
            self.groups, [{(a, b), (c, d)}, {(b, a), (d, c)}]
        )

    def pairs(self) -> Generator[list[tuple[str, str]], None, None]:
        for v1, v2 in perm2(list(self.v2e.keys())):
            if v1 == self.const or v2 == self.const:
                continue
            yield v1, v2

    def modulo(self, e: dict[str, float]) -> dict[str, float]:
        return strip(e)

    def get_all_eqs(
        self,
    ) -> dict[tuple[tuple[str, float], ...], list[tuple[str, str]]]:
        h2pairs = defaultdict(list)
        for v1, v2 in self.pairs():
            e1, e2 = self.v2e[v1], self.v2e[v2]
            e12 = minus(e1, e2)
            h12 = hashed(self.modulo(e12))
            h2pairs[h12].append((v1, v2))
        return h2pairs

    def get_all_eqs_and_why(
        self, return_quads: bool = True
    ) -> Generator[Any, None, None]:
        """Check all 4/3/2-permutations for new equalities."""
        groups = []

        for h, vv in self.get_all_eqs().items():
            if h == ():
                for v1, v2 in vv:
                    if (v1, v2) in self.eqs or (v2, v1) in self.eqs:
                        continue
                    self.eqs.add((v1, v2))
                    # why v1 - v2 = e12 ?  (note modulo(e12) == 0)
                    why_dict = minus({v1: 1, v2: -1}, minus(self.v2e[v1], self.v2e[v2]))
                    yield v1, v2, self.why(why_dict)
                continue

            if len(h) == 1 and h[0][0] == self.const:
                for v1, v2 in vv:
                    frac = h[0][1]
                    if (v1, v2, frac) in self.eqs:
                        continue
                    self.eqs.add((v1, v2, frac))
                    # why v1 - v2 = e12 ?  (note modulo(e12) == 0)
                    why_dict = minus({v1: 1, v2: -1}, minus(self.v2e[v1], self.v2e[v2]))
                    value = simplify(frac.numerator, frac.denominator)
                    yield v1, v2, value, self.why(why_dict)
                continue

            groups.append(vv)

        if not return_quads:
            return

        self.groups, links, _ = update_groups(self.groups, groups)
        for (v1, v2), (v3, v4) in links:
            if self.check_record_eq(v1, v2, v3, v4):
                continue
            e12 = minus(self.v2e[v1], self.v2e[v2])
            e34 = minus(self.v2e[v3], self.v2e[v4])

            why_dict = minus(  # why (v1-v2)-(v3-v4)=e12-e34?
                minus({v1: 1, v2: -1}, {v3: 1, v4: -1}), minus(e12, e34)
            )
            self.record_eq(v1, v2, v3, v4)
            yield v1, v2, v3, v4, self.why(why_dict)


class GeometricTable(Table):
    """Abstract class representing the coefficient matrix (table) A."""

    def __init__(self, name: str = ""):
        super().__init__(name)
        self.v2obj = {}

    def get_name(self, objs: list[Any]) -> list[str]:
        self.v2obj.update({o.name: o for o in objs})
        return [o.name for o in objs]

    def map2obj(self, names: list[str]) -> list[Any]:
        return [self.v2obj[n] for n in names]

    def get_all_eqs_and_why(self, return_quads: bool) -> Generator[Any, None, None]:
        for out in super().get_all_eqs_and_why(return_quads):
            if len(out) == 3:
                x, y, why = out
                x, y = self.map2obj([x, y])
                yield x, y, why
            if len(out) == 4:
                x, y, f, why = out
                x, y = self.map2obj([x, y])
                yield x, y, f, why
            if len(out) == 5:
                a, b, x, y, why = out
                a, b, x, y = self.map2obj([a, b, x, y])
                yield a, b, x, y, why


class RatioTable(GeometricTable):
    """Coefficient matrix A for log(distance)."""

    def __init__(self, name: str = ""):
        name = name or "1"
        super().__init__(name)
        self.one = self.const

    def add_eq(self, l1: gm.Length, l2: gm.Length, dep: pr.Dependency) -> None:
        l1, l2 = self.get_name([l1, l2])
        return super().add_eq3(l1, l2, 0.0, dep)

    def add_const_ratio(
        self, l1: gm.Length, l2: gm.Length, m: float, n: float, dep: pr.Dependency
    ) -> None:
        l1, l2 = self.get_name([l1, l2])
        return super().add_eq2(l1, l2, m, n, dep)

    def add_eqratio(
        self,
        l1: gm.Length,
        l2: gm.Length,
        l3: gm.Length,
        l4: gm.Length,
        dep: pr.Dependency,
    ) -> None:
        l1, l2, l3, l4 = self.get_name([l1, l2, l3, l4])
        return self.add_eq4(l1, l2, l3, l4, dep)

    def get_all_eqs_and_why(self) -> Generator[Any, None, None]:
        return super().get_all_eqs_and_why(True)


class AngleTable(GeometricTable):
    """Coefficient matrix A for slope(direction)."""

    def __init__(self, name: str = ""):
        name = name or "pi"
        super().__init__(name)
        self.pi = self.const

    def modulo(self, e: dict[str, float]) -> dict[str, float]:
        e = strip(e)
        if self.pi not in e:
            return super().modulo(e)

        e[self.pi] = e[self.pi] % 1
        return strip(e)

    def add_para(self, d1: gm.Direction, d2: gm.Direction, dep: pr.Dependency) -> None:
        return self.add_const_angle(d1, d2, 0, dep)

    def add_const_angle(
        self, d1: gm.Direction, d2: gm.Direction, ang: float, dep: pr.Dependency
    ) -> None:
        if ang and d2._obj.num > d1._obj.num:
            d1, d2 = d2, d1
            ang = 180 - ang

        d1, d2 = self.get_name([d1, d2])

        num, den = simplify(ang, 180)
        ang = frac(int(num), int(den))
        return super().add_eq3(d1, d2, ang, dep)

    def add_eqangle(
        self,
        d1: gm.Direction,
        d2: gm.Direction,
        d3: gm.Direction,
        d4: gm.Direction,
        dep: pr.Dependency,
    ) -> None:
        """Add the inequality d1-d2=d3-d4."""
        # Use string as variables.
        l1, l2, l3, l4 = [d._obj.num for d in [d1, d2, d3, d4]]
        d1, d2, d3, d4 = self.get_name([d1, d2, d3, d4])
        ang1 = {d1: 1, d2: -1}
        ang2 = {d3: 1, d4: -1}

        if l2 > l1:
            ang1 = plus({self.pi: 1}, ang1)
        if l4 > l3:
            ang2 = plus({self.pi: 1}, ang2)

        ang12 = minus(ang1, ang2)
        self.record_eq(d1, d2, d3, d4)
        self.record_eq(d1, d3, d2, d4)

        expr = list(ang12.items())
        if not self.add_expr(expr):
            return []

        self.register(expr, dep)

    def get_all_eqs_and_why(self) -> Generator[Any, None, None]:
        return super().get_all_eqs_and_why(True)


class DistanceTable(GeometricTable):
    """Coefficient matrix A for position(point, line)."""

    def __init__(self, name: str = ""):
        name = name or "1:1"
        self.merged = {}
        self.ratios = set()
        super().__init__(name)

    def pairs(self) -> Generator[tuple[str, str], None, None]:
        l2vs = defaultdict(list)
        for v in list(self.v2e.keys()):
            if v == self.const:
                continue
            line, p = v.split(":")
            l2vs[line].append(p)

        for line, ps in l2vs.items():
            for p1, p2 in perm2(ps):
                yield line + ":" + p1, line + ":" + p2

    def name(self, line: gm.Line, p: gm.Point) -> str:
        v = line.name + ":" + p.name
        self.v2obj[v] = (line, p)
        return v

    def map2obj(self, names: list[str]) -> list[gm.Point]:
        return [self.v2obj[n][1] for n in names]

    def add_cong(
        self,
        l12: gm.Line,
        l34: gm.Line,
        p1: gm.Point,
        p2: gm.Point,
        p3: gm.Point,
        p4: gm.Point,
        dep: pr.Dependency,
    ) -> None:
        """Add that distance between p1 and p2 (on l12) == p3 and p4 (on l34)."""
        if p2.num > p1.num:
            p1, p2 = p2, p1
        if p4.num > p3.num:
            p3, p4 = p4, p3

        p1 = self.name(l12, p1)
        p2 = self.name(l12, p2)
        p3 = self.name(l34, p3)
        p4 = self.name(l34, p4)
        return super().add_eq4(p1, p2, p3, p4, dep)

    def get_all_eqs_and_why(self) -> Generator[Any, None, None]:
        for x in super().get_all_eqs_and_why(True):
            yield x

        # Now we figure out all the const ratios.
        h2pairs = defaultdict(list)
        for v1, v2 in self.pairs():
            if (v1, v2) in self.merged:
                continue
            e1, e2 = self.v2e[v1], self.v2e[v2]
            e12 = minus(e1, e2)
            h12 = hashed(e12)
            h2pairs[h12].append((v1, v2, e12))

        for (_, vves1), (_, vves2) in perm2(list(h2pairs.items())):
            v1, v2, e12 = vves1[0]
            for v1_, v2_, _ in vves1[1:]:
                self.merged[(v1_, v2_)] = (v1, v2)

            v3, v4, e34 = vves2[0]
            for v3_, v4_, _ in vves2[1:]:
                self.merged[(v3_, v4_)] = (v3, v4)

            if (v1, v2, v3, v4) in self.ratios:
                continue

            d12 = div(e12, e34)
            if d12 is None or d12 > 1 or d12 < 0:
                continue

            self.ratios.add((v1, v2, v3, v4))
            self.ratios.add((v2, v1, v4, v3))

            n, d = d12.numerator, d12.denominator

            # (v1 - v2) * d = (v3 - v4) * n
            why_dict = minus(
                minus({v1: d, v2: -d}, {v3: n, v4: -n}),
                minus(mult(e12, d), mult(e34, n)),  # there is no modulo, so this is 0
            )

            v1, v2, v3, v4 = self.map2obj([v1, v2, v3, v4])
            yield v1, v2, v3, v4, abs(n), abs(d), self.why(why_dict)



"""Module containing the facade for the external API that must be maintained."""
# !!! Do not change the external API except if you know what you are doing !!!

from __future__ import annotations
import logging
from pathlib import Path
import traceback
from typing import Optional
from typing_extensions import Self

from newclid.auxiliary_constructions import insert_aux_to_premise
from newclid.configs import default_defs_path, default_rules_path
from newclid.graph import Graph
from newclid.numericals import draw
from newclid.ddar import solve
from newclid.problem import Problem, Theorem, Definition, Clause
from newclid.geometry import Point, Circle, Line, Segment
from newclid.write_proof import write_solution


class GeometricSolver:
    def __init__(
        self,
        proof_state: "Graph",
        problem: "Problem",
        defs: dict[str, "Definition"],
        rules: list["Theorem"],
    ) -> None:
        self.proof_state = proof_state
        self.problem = problem
        self.defs = defs
        self.rules = rules
        self.problem_string = problem.txt()

    @property
    def goal(self):
        return self.problem.goal

    def load_state(self, proof_state: "Graph"):
        self.proof_state = proof_state

    def load_problem_string(self, problem_string: str):
        self.problem_string = problem_string

    def get_problem_string(self) -> str:
        return self.problem.txt()

    def get_proof_state(self) -> str:
        return self.proof_state

    def get_defs(self):
        return self.defs

    def get_setup_string(self) -> str:
        return self.problem.setup_str_from_problem(self.defs)

    def run(self, max_steps: int = 1000) -> bool:
        solve(self.proof_state, self.rules, self.problem, max_level=max_steps)
        goal = self.problem.goal
        goal_args_names = self.proof_state.names2nodes(goal.args)
        if not self.proof_state.check(goal.name, goal_args_names):
            logging.info("Solver failed to solve the problem.")
            return False
        logging.info("Solved.")
        return True

    def write_solution(self, out_file: Path):
        write_solution(self.proof_state, self.problem, out_file)

    def draw_figure(self, out_file: Path):
        draw(
            self.proof_state.type2nodes[Point],
            self.proof_state.type2nodes[Line],
            self.proof_state.type2nodes[Circle],
            self.proof_state.type2nodes[Segment],
            block=False,
            save_to=out_file,
        )

    def get_existing_points(self) -> list[str]:
        return [p.name for p in self.proof_state.all_points()]

    def validate_clause_txt(self, clause_txt: str):
        if clause_txt.startswith("ERROR"):
            return clause_txt
        clause = Clause.from_txt(clause_txt)
        try:
            self.proof_state.copy().add_clause(clause, 0, self.defs)
        except Exception:
            return "ERROR: " + traceback.format_exc()
        return clause_txt

    def add_auxiliary_construction(self, aux_string: str):
        # Update the constructive statement of the problem with the aux point:
        candidate_pstring = insert_aux_to_premise(self.problem_string, aux_string)
        logging.info('Solving: "%s"', candidate_pstring)
        p_new = Problem.from_txt(candidate_pstring)
        p_new.url = self.problem.url
        # This is the new proof state graph representation:
        g_new, _ = Graph.build_problem(p_new, self.defs)

        self.problem = p_new
        self.proof_state = g_new


class GeometricSolverBuilder:
    def __init__(self) -> None:
        self.problem = None
        self.defs = None
        self.rules = None
        self.proof_state = None

    def build(self) -> "GeometricSolver":
        if self.problem is None:
            raise ValueError("Did not load problem before building solver.")

        if self.defs is None:
            self.defs = Definition.to_dict(
                Definition.from_txt_file(default_defs_path())
            )

        if self.rules is None:
            self.rules = Theorem.from_txt_file(default_rules_path(), to_dict=True)

        if self.proof_state is None:
            self.proof_state, _ = Graph.build_problem(self.problem, self.defs)

        return GeometricSolver(self.proof_state, self.problem, self.defs, self.rules)

    def load_problem_from_file(
        self, problems_path: Path, problem_name: str, translate: bool = True
    ) -> Self:
        problems = Problem.to_dict(
            Problem.from_txt_file(problems_path, translate=translate)
        )
        if problem_name not in problems:
            raise ValueError(f"Problem name `{problem_name}` not found in `{problems}`")
        self.problem = problems[problem_name]
        return self

    def load_problem_from_txt(
        self, problem_string: str, translate: bool = True
    ) -> Self:
        self.problem = Problem.from_txt(problem_string, translate)
        return self

    def load_rules_from_file(self, rules_path: Optional[Path] = None) -> Self:
        if rules_path is None:
            rules_path = default_rules_path()
        self.rules = Theorem.from_txt_file(rules_path, to_dict=True)
        return self

    def load_defs_from_file(self, defs_path: Optional[Path] = None) -> Self:
        if defs_path is None:
            defs_path = default_defs_path()
        self.defs = Definition.from_txt_file(defs_path, to_dict=True)
        return self



from newclid.auxiliary_constructions import insert_aux_to_premise


class TestAlphaGeometry:
    def test_insert_aux_to_premise(self):
        pstring = (
            "a b c = triangle a b c; "
            "d = on_tline d b a c, on_tline d c a b "
            "? perp a d b c"
        )
        auxstring = "e = on_line e a c, on_line e b d"

        target = (
            "a b c = triangle a b c; "
            "d = on_tline d b a c, on_tline d c a b; "
            "e = on_line e a c, on_line e b d "
            "? perp a d b c"
        )
        assert insert_aux_to_premise(pstring, auxstring) == target
